@@ -2,31 +2,59 @@ from gql import gql, Client#, AIOHTTPTransport, RequestsHTTPTransport # This is
 from gql.transport.requests import RequestsHTTPTransport
 from loguru import logger
 import base64
+from numpy.core.records import array
 
-import pandas as pd
+import pandas
 import sys
 import json
 import os
 import requests
 import pytz
 import getpass
-from time import time
 
 from .fileimport import FileImport
 from .fileImportService import FileImportService
-from .automation import Automation, Scheduler
+from .automation import Automation, Schedule
 from .programming import Programming
 
 from .utils.ut_core import Defaults, Utils
 from .utils.ut_meta import Meta
+from .timeseries import TimeSeries
 
 class TechStack(Utils, Meta):
+    """
+    Initializes a Seven2one TechStack client.
 
-    def __init__(self, host:str, user:str, password:str=None, usePorts=False, copyGraphQLString:bool=None,
-        logLevel:str='WARNING'):
-        
-        logger.remove()
-        logger.add(sys.stderr, format="{level:<10} {time} {message}", level=logLevel, colorize=True)
+    Parameters:
+    ----------
+    host: str
+        The host domain or ip address, e.g. 'app.organisation.com'
+    user: str
+        A username
+    password: str=None
+        The password belonging to the user. If None, it needs to
+        be entered interactively.
+    logLevel: str='WARNING'
+        Change the highest log level. Available log levels are 
+        'ERROR', 'WARNING', 'INFO' and 'DEBUG'.
+    usePorts: bool=False
+        A developer feature: if True, ports with prefix '8' for a
+        developer ComposeStack will be used.
+    copyGraphQLString: bool=False
+        A developer feature: if True, with each method execution, 
+        the graphQL string is copied to the clipboard (Windows only).
+
+    Examples:
+    >>> client = TechStack('app.orga.com/', 'admin', logLevel='INFO')
+    """
+
+    def __init__(self, host:str, user:str, password:str=None, logLevel:str='WARNING', 
+        usePorts:bool=False, copyGraphQLString:bool=False) -> object:
+               
+        try:
+            logger.remove()
+            logger.add(sys.stderr, format="{level:<10} {time} {message}", level=logLevel, colorize=True)
+        except: pass
 
         if os.name == 'nt': 
             logger.debug('Dectected Windows, enabling pyperclip')
@@ -43,6 +71,7 @@ class TechStack(Utils, Meta):
             automationEndpoint = f'https://{host}/automation/graphql/'
             scheduleEndpoint = f'https://{host}/schedule/graphql/'
             programmingEndpoint = f'https://{host}/programming/graphql/'
+            tsGatewayEndpoint = f'https://{host}/timeseries/graphql/'
             #fileImportServiceEndpoint = f'https://{host}/programming/graphql/'
         else: 
             tokenUrl = f'http://{host}:8040/connect/token'
@@ -50,6 +79,7 @@ class TechStack(Utils, Meta):
             automationEndpoint = f'http://{host}:8120/graphql/'
             scheduleEndpoint = f'http://{host}:8130/graphql/'
             programmingEndpoint = f'http://{host}:8140/graphql/'
+            tsGatewayEndpoint = f'http://{host}:8195/graphql/'
             fileImportServiceEndpoint = f'http://{host}:7200/graphql/'
 
         body = {
@@ -60,13 +90,14 @@ class TechStack(Utils, Meta):
             'scope': 'dynamicobjectservice'
         }
 
-        response = requests.post(tokenUrl, body)
+        response = requests.post(tokenUrl, body, verify=True)
         try:
             response =  json.loads(response.text)
             token = response['access_token']
             logger.debug(f"Got Token: {token[:10]}...")
         except:
             logger.error(response)
+            return
 
         header = {
             'authorization': 'Bearer ' + token
@@ -79,54 +110,82 @@ class TechStack(Utils, Meta):
 
         # Defaults:
         Utils._getDefaults()
-        if copyGraphQLString != None: Defaults.copyGraphQLString = copyGraphQLString
+        Defaults.copyGraphQLString = copyGraphQLString
             
-
         # self.timeZone = defaults['timeZone']
         # self.dateTimeFormat = defaults['dateTimeFormat']
 
-        self.FileImport = FileImport(self.client)
+        try:
+            self.TimeSeries = TimeSeries(self.accessToken, tsGatewayEndpoint, self)
+            self.FileImport = FileImport(self, self.TimeSeries)
+        except Exception as err:
+            logger.warning(f"Time series gateway not available")
+            logger.debug(f"Reason: {err}")
+            self.FileImport = FileImport(self)
+
         try:
             self.FileImportService = FileImportService(fileImportServiceEndpoint)
         except Exception as err:
-            logger.warning(f"No FileImportService available - {err}")
-        self.Automation = Automation(self.accessToken, automationEndpoint)
-        self.Scheduler = Scheduler(self.accessToken, scheduleEndpoint)
-        self.Programming = Programming(self.accessToken, programmingEndpoint)
-  
+            logger.debug(f"File import service not available")
+            logger.debug(f"Reason: {err}")
+        
+        try:
+            self.Automation = Automation(self.accessToken, automationEndpoint)
+        except Exception as err:
+            logger.warning(f"Automation service not available")
+            logger.debug(f"Reason: {err}")
+        
+        try:
+            self.Schedule = Schedule(self.accessToken, scheduleEndpoint)
+        except Exception as err:
+            logger.warning(f"Schedule service not available")
+            logger.debug(f"Reason: {err}")
+
+        try:
+            self.Programming = Programming(self.accessToken, programmingEndpoint)
+        except Exception as err:
+            logger.warning(f"Programming service not available")
+            logger.debug(f"Reason: {err}")
+
         return
 
-    def inventories(self, filter=None, orderBy=None, asc=True) -> pd.DataFrame:
+    def inventories(self, fields:list=None, where:str=None, orderBy:str=None, 
+        asc:bool=True) -> pandas.DataFrame:
         """
         Returns a DataFrame of existing inventories.
 
         Parameters:
         ----------
-        fields : list
-            The list of fields to be queried. Optional, defaults to 'inventoryId', 
-            'fieldName', 'displayName' and 'variant'.
-        filter : str
-            Optional, defaults to None. Use a string to add filter criteria like
-            'variant eq "BASIC"'
-        orderBy : str
-            Select a field to sort by. Defaults to None, sorting items by changedate.
-        asc : bool
-            Determines the sort order of items. Defaults to True (items sorted in 
-            ascending order). If set to False, a descending order is applied.
+        fields: list=None
+            The list of fields to be queried, e.g.
+            ['name', 'inventoryId, variant.name']
+        where: str=None
+            Use a string to add where criteria like 'name eq "meterData"'.
+            The argument is momentarily very limited.
+        orderBy: str=None
+            Select one field to sort by.
+        asc: bool=True
+            Determines the sort order of items. Set to False to apply descending order.
 
         Examples:
         >>> getInventories()
         >>> getInventories(fields=['name', 'inventoryId'], 
-                filter='fieldName contains "test"', 
+                where='city eq "Hamburg"', 
                 orderBy='variant', asc=True)
         """
 
-        fields = ['name', 'inventoryId', 'historyEnabled', 'hasValidityPeriods', 'isDomainUserType']
-        _fields = Utils._queryFields(fields)
+        if fields == None:
+            fields = ['name', 'inventoryId', 'variant.name', 'historyEnabled', 'hasValidityPeriods', 'isDomainUserType']
+            _fields = Utils._queryFields(fields, recursive=True)
+        else:
+            try:
+                _fields = Utils._queryFields(fields, recursive=True)
+            except:
+                logger.error("Fields must be provided as list, e.g. ['name', 'inventoryId, variant.name']")
 
         resolvedFilter = ''
-        if filter != None: 
-            resolvedFilter = Utils._resolveFilter(filter)
+        if where != None: 
+            resolvedFilter = Utils._resolveFilter(where)
 
         if orderBy != None:
             if asc != True:
@@ -148,133 +207,51 @@ class TechStack(Utils, Meta):
         result = Utils._executeGraphQL(self, graphQLString)
         if result == None: return
 
-        df = pd.json_normalize(result['inventories']['nodes'])
+        df = pandas.json_normalize(result['inventories']['nodes'])
         return df
 
-    def meta_items(self, inventory, pageSize:int=250):
-
-        start = time()
-
-        # unwantedColumns = ['pageInfo.hasNextPage','pageInfo.hasPreviousPage','pageInfo.startCursor', 
-        # 'pageInfo.endCursor', 'edges.cursor']
-        inventoryItemId = ''
-        allRows = []
-        cursorParam = {
-                        'name': f'{inventory}_pageSize',
-                        'value': {pageSize}
-                    }
-        count = 0
-
-        print(f"Preparation: {round(time()-start, 2)}")
-        start = time()
-
-        while True:
-            graphQLString= Meta.meta
-
-            params = {
-                'topLevelField': inventory,
-                'nestingLevel': 3,
-                'topLevelType': '',
-
-                'variables': [
-                    {
-                        'name': f'{inventory}_pageSize',
-                        'value': {pageSize}
-                    }
-                ]   
-            }
-
-            query = gql(graphQLString)
-            result = self.client.execute(query, variable_values=params)
-            print(result)
-            try:
-                result = self.client.execute(query, variable_values=params)
-                print(result)
-            except Exception as err: 
-                logger.error(err)
-                break
-
-            if count == 0:
-                Utils._copyGraphQLString(graphQLString)
-                params['variables'].append(cursorParam)
-            count += 1
-
-            rows = result['meta']['query']['dynamicItem']['result']['grid']['rows']
-            #cursor = rows[0]['values'][3]['stringValue']
-
-            allRows += rows
-
-            if len(rows) != pageSize: break
-            
-        columns = result['meta']['query']['dynamicItem']['result']['grid']['columns']
-
-        data = []
-
-        for row in allRows:
-            _row = []
-            for i in row['values']:
-                for j in i.values():
-                    _row.append(j)
-            data.append(_row)
-
-        cols = [column['path'] for column in columns]
-
-        df = pd.DataFrame(data, columns=cols)
-
-        # try:
-        #     df = df.drop(unwantedColumns, axis=1)
-        # except:
-        #     print("Warning: deletion of paging columns failed.")
-        #     return df
-
-        print(f"Transform result to Dataframe: {round(time()-start, 2)}")
-        return df
-
-    def items(self, inventoryName:str, references=False, fields:list=None, filter:str=None, pageSize:int=500,
-        orderBy=None, asc=True) -> pd.DataFrame:
+    def items(self, inventoryName:str, references:bool=False, fields:list=None, where:str=None, 
+        pageSize:int=5000, orderBy:bool=None, asc:bool=True) -> pandas.DataFrame:
         """
-        Queries items of all inventory variants. Provide a list of properties you want to 
-        retrieve, for referenced items or subfields use sublists. A sublist must be 
-        placed right behind the node.
+        Returns items of an inventory in a DataFrame.
 
         Parameters:
         -----------
         inventoryName : str
             The name of the inventory.
-        fields : list
-            A list of all properties you want to query. Defaults to all poperties.
+        references : bool
+            If True, items of referenced inventories will be added to the DataFrame. If
+            the fields-parameter is used, this parameter is ignored.
+        fields : list | str
+            A list of all properties to be queried. If None, all properties will be queried.
+            For referenced items use a '.' between inventory name and property.
         pageSize : int
             The page ize of items that is used to retrieve a large number of items.
-        filter : str
-            Use a string to add filter criteria like
+        where : str
+            Use a string to add where criteria like
             'method eq "average" and location contains "Berlin"'.
-            Defaults to None. 
+            Referenced items are not supported.
         orderBy : str
-            Select a field to sort by. Defaults to None.
+            Select a field to sort by.
         asc : bool
-            Determines the sort order of items. Defaults to True (items sorted in 
-            ascending order). If set to False, a descending order is applied.
+            Determines the sort order of items. If set to False, a descending order 
+            is applied.
 
-        Example 1 (Time Series items):
-        >>> getItems('MeterData', 
-                ['att1', 'att2', 'att3', 'resolution', ['timeUnit', 'factor'], 'unit'], 
-                filter='att2 eq "L3"')
-                )
-
-        Example 2 (default items with order):
-        >>> getItems('defaultTest', ['inventoryItemId', 'changeDate', 'orderNr', 'dateTime'], 
-                filter='orderNr lte 150 and orderNr gt 30', orderBy='orderNr', asc=False,
-                resultAsDict=False)
+        Example:
+        --------
+        >>> items('appartments', references=True)
         """
 
 
         # tz = globalTimeZone
 
         if fields != None:
+            if type(fields) != list:
+                fields = [fields]
             _fields = Utils._queryFields(fields, recursive=True)
         else:
             properties = Utils._properties(self, inventoryName, recursive=True)
-            properties = Utils._propertyList(properties)
+            properties = Utils._propertyList(properties, recursive=references)
             
             _fields = Utils._queryFields(properties, recursive=references)
 
@@ -283,8 +260,8 @@ class TechStack(Utils, Meta):
             return
 
         resolvedFilter = ''
-        if filter != None: 
-            resolvedFilter = Utils._resolveFilter(filter)
+        if where != None: 
+            resolvedFilter = Utils._resolveFilter(where)
         
         if orderBy != None:
             if asc != True:
@@ -331,7 +308,7 @@ class TechStack(Utils, Meta):
             except: 
                 break
 
-        df = pd.json_normalize(result)
+        df = pandas.json_normalize(result)
         # Remove cursor columns and remove 'node' prefix
         try:
             del df['cursor']
@@ -340,20 +317,26 @@ class TechStack(Utils, Meta):
         cols = [col.replace('node.','') for col in df.columns]
         df.columns = cols
         
-        # for key, value in properties.items():
-        #     if key in df.columns:
-        #         if value == 'DateTime':
-        #             df[key] = pd.to_datetime(df[key], format='%Y-%m-%dT%H:%M:%S').dt.tz_convert(tz)
-        #         # if globalDateTimeFormat == 'dateTime': # Need to be fixed
-        #         #     df[key] = df[key].dt.tz_localize(tz=None)
-        # if 'changeDate' in df.columns:
-        #     df['changeDate'] = pd.to_datetime(df['changeDate'], format='%Y-%m-%dT%H:%M:%S').dt.tz_convert(tz)
-        #     if globalDateTimeFormat == 'dateTime':
-        #         df['changeDate'] = df['changeDate'].dt.tz_localize(tz=None)
 
         return df
 
-    def inventoryProperties(self, name, namesOnly=False):
+    def inventoryProperties(self, inventoryName, namesOnly=False):
+        """
+        Returns a DataFrame of a query of properties of an inventory.
+
+        Parameters:
+        ----------
+        inventoryName : str
+            The name of the inventory.
+        namesOnly : bool
+            If True, only property names will be returned
+
+        Example:
+        --------
+        >>> inventoryProperties('appartments') 
+
+
+        """
 
         propertyFields = f'''
             name
@@ -367,7 +350,7 @@ class TechStack(Utils, Meta):
 
         graphQLString= f'''query Inventory {{
         inventory
-            (inventoryName: "{name}")
+            (inventoryName: "{inventoryName}")
             {{
             properties {{
                 {propertyFields}
@@ -385,44 +368,60 @@ class TechStack(Utils, Meta):
         result = Utils._executeGraphQL(self, graphQLString)
         if result == None: return
 
-        df = pd.json_normalize(result['inventory']['properties'])
+        df = pandas.json_normalize(result['inventory']['properties'])
 
         if namesOnly == True:
             return list(df['name'])
         else:
             return df
 
-    def propertyList(self, name, references=True, dataTypes=False):
+    def propertyList(self, inventoryName, references=False, dataTypes=False):
         """
-        Gets a list of properties of an inventory and its referenced inventories.
+        Returns a list of properties of an inventory and its referenced inventories
+        by reading out the scheme.
+
+        Parameters:
+        ----------
+        inventoryName : str
+            The name of the inventory.
+        references : bool
+            If True, properties of referenced inventories included.
+        dataTypes : bool
+            If True, result will be displayed as Series with properties as index and
+            dataTypes as values.
+
+        Example:
+        --------
+        >>> inventoryList('appartments') 
+
         """
 
         if references != True:
-            _properties = Utils._properties(self, inventoryName=name)
+            _properties = Utils._properties(self, inventoryName=inventoryName)
         else:
-            _properties = Utils._properties(self, inventoryName=name, recursive=True)
+            _properties = Utils._properties(self, inventoryName=inventoryName, recursive=True)
 
         if dataTypes == False:
             properties = Utils._propertyList(_properties)
         else:
-            properties = pd.Series(Utils._propertyTypes(_properties))
+            properties = pandas.Series(Utils._propertyTypes(_properties))
 
         return properties
 
-    def addBasicItems(self, inventoryName:str, items:list) -> str:
+    def addItems(self, inventoryName:str, items:list) -> str:
         """
-        Adds from a list of dicts new basic items and returns a dataframe
+        Adds from a list of dicts new items and returns a list
         of inventoryItemIds.
 
         Parameters:
         -----------
-        inventoryFieldName : str
-            The field name of the inventory.
+        inventoryName : str
+            The name of the inventory.
         items : list
-            A list of dictionaries for each item.
+            A list with dictionaries for each item.
 
-        Examples:
-        ---------
+        Example:
+        --------
         >>> items = [
                 {
                 'meterId': '86IEDD99',
@@ -433,12 +432,12 @@ class TechStack(Utils, Meta):
                 'dateTime': '2020-01-07T15:41:14Z'
                 }
             ]
-        >>> addBasicItems('meterData', items)
+        >>> addItems('meterData', items)
         """
         items = Utils._propertiesToString(items)
         #_inventoryFieldName = utils._upperFirst(inventoryFieldName)
 
-        graphQLString= f'''mutation addBasicItems {{
+        graphQLString= f'''mutation addItems {{
             create{inventoryName} (input: 
                 {items}
             )
@@ -461,13 +460,69 @@ class TechStack(Utils, Meta):
             Utils._listGraphQlErrors(result, key)
         return result[key]['InventoryItems']
 
+    def updateItems(self, inventoryName:str, items:list) -> None:
+        """
+        Updates from a list of dicts existing items. The '_inventoryItemId'
+        must be passed to each item.
+
+        Parameters:
+        -----------
+        inventoryName : str
+            The name of the inventory.
+        items : list
+            A list with dictionaries for each item.
+
+        Example:
+        --------
+        >>> items = [
+                {
+                '_inventoryItemId':'118312438662692864',
+                'meterId': '86IEDD99',
+                'dateTime': '2020-01-01T05:50:59Z'
+                },
+                {
+                '_inventoryItemId':'118312438662692864',
+                'meterId': '45IXZ52',
+                'dateTime': '2020-01-07T15:41:14Z'
+                }
+            ]
+        >>> updateItems('meterData', items)
+        """
+        items = Utils._propertiesToString(items)
+        #_inventoryFieldName = utils._upperFirst(inventoryFieldName)
+
+        graphQLString= f'''mutation updateItems {{
+            update{inventoryName} (input: 
+                {items}
+            )
+            {{
+                errors {{
+                    message
+                }}
+                InventoryItems {{
+                    _inventoryItemId
+                }}
+            }}
+        }} 
+        '''
+        result = Utils._executeGraphQL(self, graphQLString)
+        if result == None: return
+
+        key = f'update{inventoryName}'
+
+        if result[key]['errors']:
+            Utils._listGraphQlErrors(result, key)
+
+        logger.info(f"{len(result[key]['InventoryItems'])} item(s) updated.")
+        return
+
     def createInventory(self, name:str,  properties:list, variant:str=None, 
         propertyUniqueness:dict=None, historyEnabled:bool=False, 
         hasValitityPeriods:bool=False, isDomainUserType:bool=False) -> str:
         """
-        Creates an inventory
+        Creates a new inventory. After creation, access rights must be set to add items.
         
-        Parameter:
+        Parameters:
         ----------
         name : str
             Name of the new inventory (only alphanumeric characters allowed, 
@@ -480,27 +535,33 @@ class TechStack(Utils, Meta):
             Optional keys:
                 isArray: bool (Default = False)
                 nullable: bool (Default = True)
-                hasReference: bool (Default = False)
+                isReference: bool (Default = False)
                 inventoryId: str (mandatory if hasReference = True)
         variant : str
             The inventory variant.
         propertyUniqueness : list
-            A list of properties lists that should be 
-            unique in its combination. Items with the same property key can't be added.
+            A list of properties that should be unique in its combination. 
         historyEnabled : bool
             If True, changes in properties will be recorded in item history.
         hasValidityPeriods : bool
             If true, a validity period can be added to the item.    
 
         Example:
+        --------
         >>> propertyDefinitions = [
             {
-                'name': 'meterId',
+                'name': 'street',
                 'dataType': 'STRING',
-                'isArray': False,
-                'nullable': True,
-            }]
-        >>> createInventory('meterData', 'propertyDefinitions', 'TimeSeries') 
+                'nullable': False,
+            },
+            {
+                'name': 'postCode',
+                'dataType': 'STRING',
+                'nullable': False,
+            },
+            ]
+            uniqueness = [{'uniqueKey': 'address', 'properties': ['street', 'postCode']}
+        >>> createInventory('appartment', 'propertyDefinitions', propertyUniqueness=uniqueness) 
         """
 
         _properties = Utils._propertiesToString(properties)
@@ -552,16 +613,18 @@ class TechStack(Utils, Meta):
         if result['createInventory']['errors']:
             Utils._listGraphQlErrors(result, 'createInventory')
             return
+
+        logger.info(f"New inventory {name} created.")
  
         return result['createInventory']['inventory']['inventoryId']
 
-    def deleteInventory(self, name:str, force:bool=False) -> str:
+    def deleteInventory(self, inventoryName:str, force:bool=False) -> str:
         """ 
         Deletes an inventory with all its containg items. 
         
         Parameters:
         -----------
-        inventoryFieldName : str
+        inventoryName : str
             The field name of the inventory.
         force : bool
             Use True to ignore confirmation.
@@ -571,15 +634,15 @@ class TechStack(Utils, Meta):
         >>> deleteInventory('meterData', force=True)
         """
 
-        inventory = self.inventories(filter=f'name eq "{name}"')
+        inventory = self.inventories(where=f'name eq "{inventoryName}"')
         
         if inventory.empty:
-            raise Exception(f'Unknown inventory "{name}".')
+            logger.error(f'Unknown inventory "{inventoryName}".')
         inventoryId = inventory.loc[0, 'inventoryId']
         logger.debug(f'_inventoryId: {inventoryId}')
 
         if force == False:
-            confirm = input(f"Press 'y' to delete '{name}': ")
+            confirm = input(f"Press 'y' to delete '{inventoryName}': ")
 
         graphQLString= f'''mutation deleteInventory {{
             deleteInventory (input:{{inventoryId: "{inventoryId}"}})
@@ -596,13 +659,13 @@ class TechStack(Utils, Meta):
         if confirm == 'y':
             result = Utils._executeGraphQL(self, graphQLString)
             if result == None: return
-            if result['deleteInventory']['errors'] != 'None':
+            if result['deleteInventory']['errors']:
                 Utils._listGraphQlErrors(result, 'deleteInventory')
             else: 
-                logger.info(f"Inventory '{name}' successfully deleted.")
+                logger.info(f"Inventory '{inventoryName}' deleted.")
         else: return
 
-    def variants(self) -> pd.DataFrame:
+    def variants(self) -> pandas.DataFrame:
         """
             Returns a dataframe of available variants.
         """
@@ -621,47 +684,99 @@ class TechStack(Utils, Meta):
         result = Utils._executeGraphQL(self, graphQLString)
         if result == None: return
 
-        df = pd.json_normalize(result['variants']['nodes'])
+        df = pandas.json_normalize(result['variants']['nodes'])
         return df
 
-    def deleteItems(self, inventoryName:str, inventoryItemIds:list=None, filter=None, force=False):
+    def clearInventory(self, inventoryName:str, force=False):
         """
-        Deletes inventory items from a list of inventoryItemIds or from a filter. 
+        Deletes all items from the inventory
 
-        Parameters:
+        Parameters
         -----------
         inventoryName : str
             The name of the inventory.
-        items : list
-            A list of inventoryItemIds that should be deleted.
-        filter : str
-            Filter criteria to select items that should be deleted
         force : bool
             Use True to ignore confirmation.
-
-        Examples:
-        ---------
-        >>> deleteItems('meterData', filter='changeDate gt "2020-12-01"', force=True)
-        >>> deleteItems('meterData', items=['ef73d6d5-d1d7-459a-b079-ec640cbb310e'])
         """
 
-        if inventoryItemIds == None and filter == None:
-            logger.error(f"No list of items and no filter were provided.")
-        if inventoryItemIds != None and filter != None:
-            logger.warning(f"List of items and filter were provided. Item list is used.")
+        _result =self.items(inventoryName, ['inventoryItemId', 'inventoryId'])
+        _inventoryItemIds = list(_result['_inventoryItemId'])
 
-        if filter != None:
+        _ids = ''
+        for id in _inventoryItemIds:
+            _ids += f'{{_inventoryItemId: "{id}"}}\n'    
+      
+        logger.debug(f"GraphQL Ids: {_ids}")
+            
+        if force == False:
+            confirm = input(f"Press 'y' to delete  {len(_inventoryItemIds)} items: ")
+
+        graphQLString= f'''
+            mutation deleteItems {{
+                delete{inventoryName} ( input: 
+                    [{_ids}]
+                )
+                {{
+                errors {{
+                    code
+                    message
+                    }}
+                }}
+            }}
+            '''
+
+        if force == True: confirm = 'y'
+        if confirm == 'y':
+            result = Utils._executeGraphQL(self, graphQLString)
+            if result == None: return
+        else:
+            return
+
+        if result[f'delete{inventoryName}']['errors'] != None:
+            Utils._listGraphQlErrors(result, f'delete{inventoryName}')
+        else:
+            logger.info(f"{len(_inventoryItemIds)} items deleted.")
+
+    def deleteItems(self, inventoryName:str, inventoryItemIds:list=None, where:str=None, force=False):
+        """
+        Deletes inventory items from a list of inventoryItemIds or by where-criteria. 
+
+        Parameters:
+        -----------
+        inventoryName : str
+            The name of the inventory.
+        items : list
+            A list of inventoryItemIds that should be deleted.
+        where : str
+            Filter criteria to select items that should be deleted.
+            Referenced items are not supported.
+        all : bool
+            All items of the inventory will be deleted
+        force : bool
+            Use True to ignore confirmation.
+
+        Examples:
+        ---------
+        >>> deleteItems('meterData', where='changeDate gt "2020-12-01"', force=True)
+        >>> deleteItems('meterData', inventoryItemIds=['ef73d6d5-d1d7-459a-b079-ec640cbb310e'])
+        """
+
+        if inventoryItemIds == None and where == None:
+            logger.error(f"No list of items and no where-criteria were provided.")
+        if inventoryItemIds != None and where != None:
+            logger.warning(f"List of items and where-criteria has been provided. Item list is used.")
+
+        if where != None:
             _result =self.items(inventoryName, ['inventoryItemId', 'inventoryId'], 
-                filter=filter)
+                where=where)
             if _result.empty:
-                logger.info(f"The filter criteria '{filter}' led to no results.")
+                logger.info(f"The where criteria '{where}' led to no results.")
                 return
 
             _inventoryItemIds = list(_result['_inventoryItemId'])
             
         if inventoryItemIds != None:
             _inventoryItemIds = inventoryItemIds
-            #_inventoryItemIds = json.dumps(items)
 
         _ids = ''
         for id in _inventoryItemIds:
@@ -669,7 +784,6 @@ class TechStack(Utils, Meta):
       
         logger.debug(f"GraphQL Ids: {_ids}")
             
-        
         if force == False:
             confirm = input(f"Press 'y' to delete  {len(_inventoryItemIds)} items: ")
 
@@ -700,8 +814,190 @@ class TechStack(Utils, Meta):
         else:
             logger.info(f"{len(_inventoryItemIds)} items deleted.")
 
+    def updateVariant(self, variantName, newName=None, icon=None) -> None:
+        """Updates a variant"""
 
+        _variantId = Utils._getVariantId(self.variants(), variantName) 
+        logger.debug(f"Found variantId: {_variantId}")
+        if type(_variantId) != str:
+            logger.er
+
+        if newName != None:
+            _name = f'name: "{newName}"'
+        else:
+            _name = ''
+
+        if icon != None:
+            _icon =f'icon: "{icon}"'
+        else:
+            _icon =''
+
+        graphQLString = f'''mutation updateVariant {{
+            updateVariant (input:{{
+                variantId: "{_variantId}"
+                {_name}
+                {_icon}
+                }})
+            {{ 
+                errors {{ 
+                    code
+                    message
+                }}
+            }}
+            }}
+            '''
+
+        result = Utils._executeGraphQL(self, graphQLString)
+        if result == None: return
+        print(result)
+        # if result['updateVariant']['errors'] != 'None':
+        #     Utils._listGraphQlErrors(result, 'updateVariant')
+        # else: 
+        #     logger.info(f"Variant '{variantName}' updated.")
+
+        return
+
+    def updateArrayProperty(self, inventoryName:str, inventoryItemId:str, arrayProperty:str, 
+        operation:str, arrayItems:list=None, cascadeDelete:bool=False) -> None:
+        """
+        Updates a single array property of a single inventoryItemId. Arrays with and without 
+        references are supported.
+
+        Parameters:
+        -----------
+        inventoryName : str
+            The name of the inventory where the item is located.
+        inventoryItemId : str
+            The inventory item id of the item.
+        arrayProperty : str
+            The name of the property whose array items are to be updated.
+        operation : str
+            The update operation to be performed. Options:
+            insert: inserts a list of array items.
+            removeById: removes array items by a list of given (item) ids.
+            removeByIndex: removes array items by a list of given indices.
+            removeAll: removes all array items
+        cascadeDelete : bool (False)
+            If array items are refencences, True will delete also the reference items. 
+
+        Examples:
+        ---------
+        >>> client.updateArrayProperty('meterData', '118718694909018112',
+                'timeSeries', action='insert', arrayItems=['387964598252380179'])
+        >>> client.updateArrayProperty('meterData', '118718694909018112',
+                'timeSeries', action='removeByIndex', arrayItems=[0,1], cascadeDelete=True)
+        """
+
+
+        operations = ['insert', 'removeById', 'removeByIndex', 'removeAll']
+        if operation not in operations:
+            logger.error(f"Action '{operation}' allowed. Possible update operations: {operations} ")
+
+        if operation == 'removeAll':
+            try:
+                arrDf= TechStack.items(self, inventoryName, fields=[f'{arrayProperty}._inventoryItemId'],
+                    where=f'_inventoryItemId eq "{inventoryItemId}"')
+            except Exception as err:
+                logger.error(err)
+                return
+            countArray = len(arrDf.iloc[0,0])
+            arrayItems = [num for num in range(countArray)]
+        _arrayItems = Utils._arrayItemsToString(arrayItems, operation, cascadeDelete)
+
+        graphQLString = f'''mutation updateArray {{
+            update{inventoryName}ArrayProperties(
+                input: {{
+                _inventoryItemId: "{inventoryItemId}"
+                {arrayProperty}: {{
+                    {_arrayItems}
+                    }}
+                }}
+            ) {{
+                errors {{
+                message
+                }}
+            }}
+        }}'''
+
+        result = Utils._executeGraphQL(self, graphQLString)
+        if result == None: return
+
+        key = f'update{inventoryName}ArrayProperties'
+        if result[key]['errors'] != None:
+            Utils._listGraphQlErrors(result, key)
+        else:
+            logger.info(f"Array property {arrayProperty} for item {inventoryItemId} updated.")
+
+    def addInventoryProperties(self, inventoryName:str, properties:list) -> None:
+        """
+        Adds one or more inventory properties to an existing inventory.
+
+        Parameters:
+        ----------
+        inventoryName: str
+            Name of inventory
+        properties: list
+            A list of dicts with the following mandatory keys: 
+                name: str
+                dataType: enum (STRING, BOOLEAN, DECIMAL, INT, LONG, DATE_TIME, 
+                DATE_TIME_OFFSET)
+            Optional keys:
+                isArray: bool (Default = False)
+                nullable: bool (Default = True)
+                isReference: bool (Default = False)
+                inventoryId: str (mandatory if hasReference = True)
+        """
+
+        key = 'addProperty' 
+        inventory = self.inventories(where=f'name eq "{inventoryName}"')
+        
+        if inventory.empty:
+            logger.error(f'Unknown inventory "{inventoryName}".')
+        inventoryId = inventory.loc[0, 'inventoryId']
+        _properties = Utils._propertiesToString(properties)
+
+        graphQLString= f'''
+        mutation {key} {{
+        {key} (input: {{
+            inventoryId: "{inventoryId}"	
+            properties: {_properties}
+            }}) 
+            {{
+                errors {{
+                    message
+                }}
+            }}
+        }}
+        '''
+        result = Utils._executeGraphQL(self, graphQLString)
+        if result == None: return
+
+        if result[key]['errors']:
+            Utils._listGraphQlErrors(result, key)
+            return
+        else:
+            logger.info(f"New property(ies) added.")
+
+    def resync(self) -> None:
+        """Resynchornizes read databases"""
+
+        key = 'reSyncReadDatabase'
+        graphQLString= f'''mutation resync{{
+            {key}
+            }}
+        '''
+        result = Utils._executeGraphQL(self, graphQLString)
+        if result == None: return
+
+        return result
+    
+    def defaultDataFrame(self, maxRows, maxColumns):
+        """Sets default sizes for a DataFrame for the current session"""
+        pandas.options.display.max_rows = maxRows
+        pandas.options.display.max_columns = maxColumns
+        return
 # Some stand alone functions
+
 def setGlobalDefault(default:str, value:object, verbose:bool=False) -> None:
     """
     Sets a default value for a specific option.
@@ -775,8 +1071,6 @@ def setGlobalDefault(default:str, value:object, verbose:bool=False) -> None:
     if verbose: print(f"{default} set to {value}.")
     return
 
-
-
 def encodeBase64(file:str):
     with open(file) as file:
         content = file.read()
