@@ -1,4 +1,6 @@
 import pandas as pd
+import json
+import os
 from gql import gql, Client#, AIOHTTPTransport, RequestsHTTPTransport # This is gql version 3
 from gql.transport.requests import RequestsHTTPTransport
 from loguru import logger
@@ -14,8 +16,8 @@ class Programming():
             'authorization': 'Bearer ' + accessToken
         }
         
-        transport =  RequestsHTTPTransport(url=endpoint, headers=header, verify=False)
-        self.client = Client(transport=transport, fetch_schema_from_transport=True)
+        transport =  RequestsHTTPTransport(url=endpoint, headers=header, verify=True)
+        self.client = Client(transport=transport, fetch_schema_from_transport=False)
 
         return
 
@@ -38,8 +40,7 @@ class Programming():
         return df
 
     def createFunction(self, name:str, languageVersion:str, description:str=None, 
-        files:list=None, deploy:bool=False, automationTopic:str=None, 
-        envVars:dict=None) -> str:
+        files:list=None, deploy:bool=False, deploymentName:str=None, envVars:dict=None) -> str:
         """ 
         Creates a function with possibility to commit files and deployment
         in one step. The function Id is returned.
@@ -59,9 +60,9 @@ class Programming():
         deploy : bool
             If True, the function will be deployed. The files argument must not be 
             None in this case.
-        automationTopic : str
-            The connector to a Camunda workflow. If no topic is provided, the function can only
-            be executed with the Programming Service, but cannot used in an Automation workflow.
+        deploymentName:
+            The name of the deployed function. If left to default (None) the function name 
+            will be used. Use small letters only and no special characters.
         envVars : dict
             A Dictionary of Environment variables provided to the deployment function. All values
             will be converted to string.
@@ -72,7 +73,7 @@ class Programming():
         >>> files = [path1, path2, path3]
             vars = {'var1': 42, 'var2': 'full_scale'}
             createFunction('myfunction', 'PYTHON_3_9', files=files, deploy=True, 
-                automationTopic='autFunction', envVars=vars)
+                envVars=vars)
         """
 
         graphQLString = f'''mutation createFunction {{
@@ -111,11 +112,13 @@ class Programming():
             if deploy == False:
                 return functionId
             else:
-                self.deployFunction(functionId, name, automationTopic, envVars)
+                if deploymentName == None:
+                    deploymentName = name.lower()
+                self.deployFunction(functionId, deploymentName, envVars)
                 return functionId
 
     def commitFunctionFiles(self, functionId:str, files:list=None, deploy:bool=False, 
-        deploymentName:str=None, automationTopic:str=None, envVars:dict=None) -> None:
+        deploymentName:str=None, envVars:dict=None) -> None:
         """
         Upload programming files to an existing function. 
 
@@ -129,9 +132,6 @@ class Programming():
             If True, the function will be deployed, after commintg the files. 
         deploymentName:
             The name of the deployed function. Use small letters only and no special characters.
-        automationTopic : str
-            The connector to a Camunda workflow. If no topic is provided, the function can only
-            be executed with the Programming Service, but cannot used in an Automation workflow.
         envVars : dict
             A Dictionary of Environment variables provided to the deployment function. All values
             will be converted to string.
@@ -141,7 +141,7 @@ class Programming():
         >>> files = [path1, path2, path3]
             vars = {'var1': 42, 'var2': 'full_scale'}
             commitFunctionFiles('3dba276e3c8645838c2d598043cab057', files=files, deploy=True, 
-                deploymentName=myfunction, automationTopic='autFunction', envVars=vars)
+                deploymentName='myfunction', envVars=vars)
         """
 
         fileList = AutProgUtils._upsetFiles(files)
@@ -173,14 +173,11 @@ class Programming():
             else:
                 if deploymentName == None:
                     logger.error(f"No deployment name was provided for deployment. Function id: {functionId}")
-                if automationTopic==None:
-                    automationTopic = ''
-                self.deployFunction(functionId, deploymentName, automationTopic, envVars)
+                self.deployFunction(functionId, deploymentName, envVars)
 
         return result
 
-    def deployFunction(self, functionId:str, deploymentName:str, automationTopic:str=None, 
-        envVars:dict=None):
+    def deployFunction(self, functionId:str, deploymentName:str, envVars:dict=None):
         """
         Deploys a function to make it executable.
 
@@ -190,34 +187,26 @@ class Programming():
             Id of an existing function.
         deploymentName:
             The name of the deployed function. Use small letters only and no special characters.
-        automationTopic : str
-            The connector to a Camunda workflow. If no topic is provided, the function can only
-            be executed with the Programming Service, but cannot used in an Automation workflow.
         envVars : dict
-            A Dictionary of Environment variables provided to the deployment function. All values
-            will be converted to string.
+            A Dictionary of Environment variables provided to the deployment function. 
 
         Example:
         --------
-        >>> vars = {'var1': 42, 'var2': 'full_scale'}
+        >>> vars = {'exec_timeout': '2m', 'read_timeout': '2m','write_timeout': '2m'}
             deployFunction('3dba276e3c8645838c2d598043cab057', deploymentName=myfunction, 
             automationTopic='autFunction', envVars=vars)
         """
 
-        if automationTopic == None:
-            automationTopic = ''
-
         if envVars == None:
             _vars = ''
         else:
-            _vars = AutProgUtils._varsToString(vars, 'env')
+            _vars = AutProgUtils._varsToString(envVars, 'env')
 
         graphQLString = f'''mutation deployFunction {{
             deployFunction(
                 input: {{
                 functionId: "{functionId}"
                 functionName: "{deploymentName}"
-                automationTopic: "{automationTopic}"
                 {_vars}
                 }}
             ) {{
@@ -237,7 +226,7 @@ class Programming():
             logger.info(f"Deployed function with deploymentId {deploymentId}")
             return deploymentId
 
-    def functionFiles(self, functionId:str) -> pd.DataFrame:
+    def functionFiles(self, functionId:str, downloadPath:str=None) -> pd.DataFrame:
         """
         Shows files of a function.
 
@@ -251,11 +240,17 @@ class Programming():
         >>> functionFiles('3dba276e3c8645838c2d598043cab057')
         """
 
+        if downloadPath == None:
+            _contentBase64 = ''
+        else:
+            _contentBase64 = 'contentBase64'
+            
         graphQLString = f''' query functionFiles {{
             functionFiles (functionId: "{functionId}") {{
                 version
                 files {{
                     fullname
+                    {_contentBase64}
                     }}
                 }}
             }}
@@ -265,9 +260,21 @@ class Programming():
         if result == None: return
 
         df = pd.json_normalize(result['functionFiles'], meta=['version'], record_path=['files'])
+
+        if downloadPath != None:
+            if os.path.exists(downloadPath) == False:
+                logger.error(f"Download path '{downloadPath}' does not exist.")
+            else:
+                fileList = []
+                for file in result['functionFiles']['files']:
+                    AutProgUtils._downloadFunctionFile(downloadPath, file['fullname'], file['contentBase64'])
+                    fileList.append(file['fullname'])
+                logger.info(f"Downloaded {fileList} to {downloadPath}")
+                del df['contentBase64']
+        
         return df
 
-    def executeFunction(self, deploymentId:str) -> str:
+    def executeFunction(self, deploymentId:str, inputVariables:dict=None) -> str:
         """
         Executes a function and returns its execution id
         
@@ -281,12 +288,18 @@ class Programming():
         >>> executeFunction('c877cc1b568a4c489aacdb4538b3f544')
         """
 
+        if inputVariables == None:
+            _vars = ''
+        else:
+            _vars = json.dumps(inputVariables)
+            _vars = _vars.replace('"', '\\"')
 
         graphQLString = f''' mutation executeFunction {{
             executeFunction(
                 input: {{ 
                     deploymentId: "{deploymentId}", 
-                    input: "OUTPUT" }}
+                    input: "{_vars}"
+                    }}
             ) {{
                 executionId
                 result {{
