@@ -1,11 +1,12 @@
 
 import json
-import sys
+import pytz
 import os
 import base64
 import pandas as pd
 from gql import gql
 from dataclasses import dataclass
+from tzlocal import get_localzone
 from loguru import logger
 
 if os.name == 'nt': #only if the platform is Windows, pyperclip will be imported
@@ -18,6 +19,13 @@ class Defaults:
     copyGraphQLString: bool = False
 
 class Utils():
+
+    def _timeZone(timeZone):
+        if timeZone=='local': 
+            localTimeZone = get_localzone().zone
+            return str(pytz.timezone(localTimeZone))
+        else:
+            return str(pytz.timezone(timeZone))
     
     def _queryFields(fieldList:list, recursive=False):
         """
@@ -26,6 +34,7 @@ class Utils():
         """
         fields = ''
         splitList = [item.split('.') for item in fieldList]
+        logger.debug(f"Queried fields: {splitList}")
 
         def nestedItem(item):
             nonlocal fields
@@ -39,7 +48,7 @@ class Utils():
             fields += line
     
         for item in splitList:
-            
+           
             if len(item) == 1:
                 fields += f'{item[0]}  \n'
             else:
@@ -47,14 +56,15 @@ class Utils():
                 if recursive == False:
                     if item[-1] == '_displayValue':  
                         nestedItem(item)
+                    if item[-1] == '_inventoryItemId':  
+                        nestedItem(item)
                 else:
                     nestedItem(item)
-
         return fields
 
-    def _resolveFilter(filterString:str) -> str:
+    def _resolveFilter(whereString:str) -> str:
         """
-        Resolves a simple filter string into a GraphQL argument string
+        Resolves a simple where string into a GraphQL argument string
         """  
         EXECEPTION_STRING = f"""Filter syntax not correct - example syntax: 'colour eq "blue" and size gt "10"'"""
 
@@ -62,12 +72,12 @@ class Utils():
         positionList = []
         p = 0
         while p != -1:
-            p = filterString.find('"', p + 1)
+            p = whereString.find('"', p + 1)
             if p != -1: positionList.append(p)
         if len(positionList) == 0:
             p = 0
             while p != -1:
-                p = filterString.find("'", p + 1)
+                p = whereString.find("'", p + 1)
                 if p != -1: positionList.append(p)
         if len(positionList) % 2 != 0:
             raise ValueError(EXECEPTION_STRING)
@@ -83,7 +93,7 @@ class Utils():
         _components = []
         for i in range(len(positionList)):
             try:
-                part = filterString[positionList[i] : positionList[i+1]]
+                part = whereString[positionList[i] : positionList[i+1]]
                 _components.append(part) 
             except: pass
         components = []
@@ -105,7 +115,7 @@ class Utils():
         components = [i for i in components if i != '']
         components = [i.replace("'", '"') for i in components if i != '']
         
-        ## Determine the correct numer of arguments a filter may have
+        ## Determine the correct numer of arguments where-criteroa may have
         correctArgLength = [x * 4 - 1 for x in range(1, 40)]
         
         ## Convert components list into GraphQL Filter String
@@ -123,7 +133,6 @@ class Utils():
             return argOpener + argFirst + argFollow + argCloser
         else:
             raise ValueError(EXECEPTION_STRING)
-            return
     
     def _propertiesToString(properties:list) -> str:
         """ Converts a list of property dicts for many items into a string """
@@ -138,14 +147,50 @@ class Utils():
         _properties += ']'
         return _properties
 
+    def _tsPropertiesToString(properties:list) -> str:
+        """ Converts a list of property dicts for many items into a string """
+        timeUnit, factor = 'timeUnit', 'factor'
+        _properties = '[\n'
+        for property in properties:
+            _properties += '{\n'
+            for key, value in property.items():
+                if key == 'resolution':
+                    try:
+                        _properties += f'{key}: {{\n'
+                        _properties += f'timeUnit: {value[timeUnit]}\n'
+                        _properties += f'factor: {value[factor]}\n'
+                        _properties += f'}}\n'
+                    except KeyError:
+                        logger.error("Missing 'timeUnit' and/or 'factor' for Timeseries resolution")
+                        return
+                else:
+                    _properties += Utils._customProperties(key, value)
+            
+            _properties += '}\n'
+        _properties += ']'
+        return _properties
+
+    def _addToGroupPropertiesToString(groupItemId:str, properties:list) -> str:
+        """ Converts a list of property dicts for many items into a string """
+        
+        _properties = '[\n'
+        for property in properties:
+            _properties += f'{{_groupInventoryItemId: "{groupItemId}"\n'
+            for key, value in property.items():
+                _properties += Utils._customProperties(key, value)      
+            
+            _properties += '}\n'
+        _properties += ']'
+        return _properties
+
     def _uniquenessToString(propertyUniqueness: list):
         """ Converts a list of unique keys into a string """    
 
         _uniqueKeys = '[\n'
         for item in propertyUniqueness:
-            key = item['key']
-            _uniqueKeys += f'{{key: "{key}" value: ['
-            for value in item['value']:
+            key = item['uniqueKey']
+            _uniqueKeys += f'{{uniqueKey: "{key}" properties: ['
+            for value in item['properties']:
                 _uniqueKeys += f'"{value}",'     
             
             _uniqueKeys += ']}\n'
@@ -181,7 +226,7 @@ class Utils():
             _propertyString += f'{key}: {_value}]\n'
 
         else:
-            print(f"Error: {value} is an unsupported value type.")
+            logger.error(f"{value} is an unsupported value type.")
 
         return _propertyString
 
@@ -203,40 +248,56 @@ class Utils():
                 if key['name'] == dictName:
                     for property in key['fields']:
                         if property['name'] not in systemProperties:
-                            if property['type']['name'] in types:
-                                subdict.setdefault(property['name'], property['type']['name']) 
+                            if property['type']['kind'] != 'OBJECT':
+                                subdict.setdefault(property['name'], property['type']['name'])
                             else:
                                 subdict.setdefault(property['name'], makeDict(property['type']['name']))
-
+            logger.debug(f"Subdict: {subdict}")
             return subdict
 
         for key in _scheme:
             if key['name'] == inventoryName:
                 for property in key['fields']:
                     if property['name'] not in systemProperties:
-                        if property['type']['name'] in types:
-                            properties.setdefault(property['name'], property['type']['name']) 
-                            pass
+                        if property['type']['kind'] != 'OBJECT':
+                            if property['type']['kind'] == 'LIST':
+                                if property['type']['ofType']['kind'] == 'OBJECT':
+                                    properties.setdefault(property['name'], {'_inventoryItemId': 'String'})
+                                else:
+                                    properties.setdefault(property['name'], property['type']['name'])
+                            else:
+                                properties.setdefault(property['name'], property['type']['name'])
                         else:
                             if recursive == True:
-                                properties.setdefault(property['name'], makeDict(property['type']['name']))
+                                logger.debug("Recursive")
+                                if property['type']['kind'] =='LIST':
+                                    properties.setdefault(property['name'], {'_inventoryItemId': 'String'})
+                                else:
+                                    properties.setdefault(property['name'], makeDict(property['type']['name']))
                             else:
-                                properties.setdefault(property['name'], property['type']['name']) 
-
+                                properties.setdefault(property['name'], property['type']['name'])
+                
+        logger.debug(f"Properties of inventory: {properties}")
         return properties
 
-    def _propertyList(propertyDict:dict) -> list:
+    def _propertyList(propertyDict:dict, recursive:bool=False) -> list:
         """Uses _properties() to create a flat list of properties"""
 
         propertyList = []
     
         def instDict(subDict, path):
-            for k, v in subDict.items():
-                if isinstance(v, dict):
-                    path = f'{path}.{k}'
-                    instDict(subDict[k], path)
-                else:
-                    propertyList.append(f'{path}.{k}')
+            if recursive == True:
+                for k, v in subDict.items():
+                    if isinstance(v, dict):
+                        path = f'{path}.{k}'
+                        instDict(subDict[k], path)
+                    else:
+                        propertyList.append(f'{path}.{k}')
+            else:
+                if '_displayValue' in subDict.keys(): #if it is not an array
+                    propertyList.append(f'{path}._displayValue')
+                else: # if it is an array
+                    propertyList.append(f'{path}._inventoryItemId')
 
         for k, v in propertyDict.items():
             if isinstance(v, dict):
@@ -244,6 +305,7 @@ class Utils():
             else:
                 propertyList.append(k)
 
+        logger.debug(f'Property list: {propertyList}')
         return propertyList
             
     def _propertyTypes(propertyDict:dict) -> dict:
@@ -295,7 +357,13 @@ class Utils():
 
     def _getVariantId(variants:pd.DataFrame, name:str) -> str:
         """ Gets the variant Id from a given name""" 
-        return variants.where(variants['name'] == name)['variantId'][0]
+        variants.set_index('name', inplace=True)
+        return variants.loc[name][0]
+
+    def _getInventoryId(inventories:pd.DataFrame, name:str) -> str:
+        """ Gets the variant Id from a given name""" 
+        inventories.set_index('name', inplace=True)
+        return inventories.loc[name][0]
 
     def _listGraphQlErrors(result:dict, key:str) -> None:
         """Print errors from GraphQL Query to log"""
@@ -309,11 +377,44 @@ class Utils():
             content = base64.b64encode(content.encode('ascii'))
             return content.decode('UTF8')
 
+    def _arrayItemsToString(arrayItems:list, operation:str, cascadeDelete:bool) -> str:
+        """Converts a list of array items to a graphQL string"""
+
+        cDelValue = 'true' if cascadeDelete == True else 'false'
+
+        if operation == 'insert':
+            _arrayItems = 'insert: [\n'
+            for item in arrayItems:
+                _arrayItems += f'{{value: "{item}"}}\n'
+            _arrayItems += ']'
+            return _arrayItems
+        if operation == 'removeByIndex':
+            _arrayItems = f'cascadeDelete: {cDelValue}\n'
+            _arrayItems += 'removeByIndex: ['
+            for item in arrayItems:
+                _arrayItems += f'{item}, '
+            _arrayItems += ']'
+            return _arrayItems
+        if operation == 'removeById':
+            _arrayItems = f'cascadeDelete: {cDelValue}\n'
+            _arrayItems += 'removeById: ['
+            for item in arrayItems:
+                _arrayItems += f'"{item}", '
+            _arrayItems += ']'
+            return _arrayItems
+        if operation == 'removeAll':
+            _arrayItems = f'cascadeDelete: {cDelValue}\n'
+            _arrayItems += 'removeByIndex: ['
+            for item in arrayItems:
+                _arrayItems += f'{item}, '
+            _arrayItems += ']'
+            return _arrayItems
+
     def _executeGraphQL(self, graphQLString):
         """Executes GraphQl, this code is used in every main function"""
 
         Utils._copyGraphQLString(graphQLString)      
-        logger.debug(graphQLString)
+        logger.debug(f"GraphQLString: {graphQLString}")
         query= gql(graphQLString)
 
         try:
